
// Compute the dot product of long vecrtors using CUDA
// Uses a 1D grid layout; if you have huge vectors, you'll need to use a 2D layout
//
// nvcc -arch=sm_35 --compiler-options -Wall,-ansi dotprod.cu -o dotprod
//
// GZ, Feb 2013


#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#ifdef _WIN32
#define srand48(expr) srand(expr)
#define drand48() double(rand())/double(RAND_MAX)
#endif

#define THREADS_PER_BLOCK 512
// MUST be 2^k !
// this constant should be a parameter for a template kernel
// as described in http://stackoverflow.com/questions/9187899/cuda-shared-memory-array-variable



// Multiply two vectors 'a' and 'b' of length 'n', put the partial sums in c.
// Prerequisite: 'c' must hold at least n/nBlocks many floats!
// Use shared memory to improve performance

__global__
void dotprod( const float *a, const float *b, float *c, unsigned int n )
{
	__shared__ float cache[THREADS_PER_BLOCK];
	unsigned int tid = blockDim.x * blockIdx.x + threadIdx.x;
	if ( tid < n )
		cache[threadIdx.x] = a[tid] * b[tid];

	// for reduction, THREADS_PER_BLOCK must be a power of 2!
	__syncthreads();
	int i = blockDim.x/2;	// first pass: add upper half to lower half
	while ( i != 0 )
	{
		if ( threadIdx.x < i )
			cache[threadIdx.x] += cache[threadIdx.x + i];
		__syncthreads();
		i /= 2;
	}

	// one thread copies partial sum to global memory
	if ( threadIdx.x == 0 )
		c[blockIdx.x] = cache[0];
}



// General CUDA error handler
// Prints error message, if something went wrong
void handleCUDAerror( int line )
{
	cudaError_t err = cudaGetLastError();
	if ( err == cudaSuccess )
		return;

	printf("CUDA error message: %s\n"
		   "line %d\n", cudaGetErrorString(err), line );

	int devID;
    cudaGetDevice( &devID );
    cudaDeviceProp props;
    cudaGetDeviceProperties( &props, devID );
    printf("CUDA device %d: \"%s\" with compute capability %d.%d\n",
           devID, props.name, props.major, props.minor);
}



// Check whether our grid layout fits with the current GPU's maximum properties
// Warning: considers only X and Y directions !
void checkMaxProperties( dim3 blocks, dim3 threads )
{
	int devID;
    cudaGetDevice( &devID );
    cudaDeviceProp props;
    cudaGetDeviceProperties( &props, devID );
	handleCUDAerror(__LINE__);

	if ( threads.x * threads.y > unsigned(props.maxThreadsPerBlock) )
	{
		fprintf(stderr, "Too many threads per block! (%d x %d > %d)\n",
				threads.x, threads.y, props.maxThreadsPerBlock );
		exit(1);
	}
	if ( threads.x > unsigned(props.maxThreadsDim[0]) )
	{
		fprintf(stderr, "Too many threads per block along X direction! (%d > %d)\n",
				threads.x, props.maxThreadsDim[0] );
		exit(1);
	}
	if ( threads.y > unsigned(props.maxThreadsDim[1]) )
	{
		fprintf(stderr, "Too many threads per block along Y direction! (%d > %d)\n",
				threads.y, props.maxThreadsDim[1] );
		exit(1);
	}

	if ( blocks.x > unsigned(props.maxGridSize[0]) )
	{
		fprintf(stderr, "Too many blocks in grid along X direction! (%d > %d)\n",
				blocks.x, props.maxGridSize[0] );
		exit(1);
	}
	if ( blocks.y > unsigned(props.maxGridSize[1]) )
	{
		fprintf(stderr, "Too many blocks in grid along Y direction! (%d > %d)\n",
				blocks.y, props.maxGridSize[1] );
		exit(1);
	}

	// all's well ...
}



// Round up to the next highest power of 2
// From: http://graphics.stanford.edu/~seander/bithacks.html
// Details: result may be expressed by the formula 1U << (lg(v - 1) + 1)
// Warning: if v==0, then returns 0 (not 1)!
unsigned int roundUpToPowerOf2( unsigned int v )
{
	v--;
	v |= v >> 1;
	v |= v >> 2;
	v |= v >> 4;
	v |= v >> 8;
	v |= v >> 16;
	v++;

	return v;
}


int main( int argc, char **argv )
{
	unsigned int vec_len = 512;									// just a default
	if ( argc > 1 )
		// interpret first argument as length
		vec_len = atoi( argv[1] );

	// vector passing
	// make vec_len a power of 2 (so our simple reduction algo in the kernel works)
	if ( vec_len < THREADS_PER_BLOCK )
		vec_len = THREADS_PER_BLOCK;
	vec_len = roundUpToPowerOf2( vec_len );
	printf( "Multiply two vectors of length %d\n", vec_len );

	// set layout of blocks and threads; we make a 1D threads/blocks layout here
	unsigned int num_blocks = (vec_len + THREADS_PER_BLOCK - 1) / THREADS_PER_BLOCK;
	// this funny expression yields ceil( max_blocks_per_grid / threads_per_block ),
	// but needs only integer arithmetic
	printf( "Threads per block = %d\n", THREADS_PER_BLOCK );
	printf( "Number of blocks  = %d\n", num_blocks );
	dim3 threads_per_block( THREADS_PER_BLOCK, 1 );
	dim3 n_blocks( num_blocks, 1);
	checkMaxProperties( n_blocks, threads_per_block );

	// fill two vectors with random numbers
	size_t in_size = vec_len * sizeof(float);
	float * h_a = (float*) malloc( in_size );
	float * h_b = (float*) malloc( in_size );
	float out_size = num_blocks * sizeof(float);
	float * h_c = (float*) malloc( out_size );
	if ( h_a == NULL || h_b == NULL || h_c == NULL )
    {
        fprintf(stderr, "Failed to allocate host vectors!\n");
		exit(1);
	}

	srand48( time(NULL) );
    for ( unsigned int i = 0; i < vec_len; i++ )
    {
        h_a[i] = drand48();
        h_b[i] = drand48();
    }
	
	// transfer the arrays to the device
	float *d_a, *d_b, *d_c;
	cudaMalloc( (void **) &d_a, in_size );
	cudaMalloc( (void **) &d_b, in_size );
	cudaMalloc( (void **) &d_c, out_size );
	handleCUDAerror(__LINE__);

	cudaMemcpy( d_a, h_a, in_size, cudaMemcpyHostToDevice );
	cudaMemcpy( d_b, h_b, in_size, cudaMemcpyHostToDevice );
	handleCUDAerror(__LINE__);						// yes, we should do this after every memcpy ...

	// launch the vector multiplication kernel
	dotprod<<< n_blocks, threads_per_block >>>( d_a, d_b, d_c, vec_len );
	handleCUDAerror(__LINE__);

	// copy partial results back to the host (does cudaDeviceSynchronize implicitely)
	cudaMemcpy( h_c, d_c, out_size, cudaMemcpyDeviceToHost );
	handleCUDAerror(__LINE__);

/*
for ( unsigned int i = 0; i < num_blocks; i ++ )
  printf("%f ",  h_c[i]);
putchar('\n');
for ( unsigned int i = 0; i < vec_len; i++ )
{
double p= 0.0;
  if ( i % THREADS_PER_BLOCK == 0 )
    p= 0.0;
  p +=  h_a[i] * h_b[i];
  if ( i % THREADS_PER_BLOCK == THREADS_PER_BLOCK-1 )
    printf("partial-sum %d = %f\n", (i / THREADS_PER_BLOCK), p );
}
 */ 

	// compute rest on CPU
	float prod_device = 0.0;
	for ( unsigned int i = 0; i < num_blocks; i ++ )
		prod_device += h_c[i];

	// verify the results (kind of a unit test)
	bool verify_ok = true;
	double prod_host = 0.0;
	for ( unsigned int i = 0; i < vec_len; i++ )
		prod_host += h_a[i] * h_b[i];
	printf("Difference (dotprod on host) - (dotprod on device) = %f - %f = %f\n",
			prod_host, prod_device, (prod_host - prod_device) );
	if ( fabsf(prod_host - prod_device) > vec_len*1e-7 )
	{
		printf("Error: difference larger than %f!\n", vec_len*1e-7 );
		verify_ok = false;
	}

	// free all memory
	cudaFree( d_a );
	cudaFree( d_b );
	cudaFree( d_c );
	handleCUDAerror(__LINE__);

	free( h_a );
	free( h_b );
	free( h_c );

	cudaDeviceReset();
	if ( verify_ok )
		puts("Success.");
	return 0;
}

